/*

Htpdate v0.2

Extract date/time stamp from web server response for time synchronization.
This program works with the most common date/time stamps format specified in
RFC 822 updated by RFC 1123.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.
http://www.gnu.org/copyleft/gpl.html

Version history
v0.1	Eddy Vervest
v0.2	Roy Keene

*/


#include <stdio.h>			/* Needed for printf() */
#include <string.h>			/* Needed for strstr() and strcpy() */
#include <netinet/in.h>		/* Needed for internet address structure. */
#include <sys/socket.h>		/* Needed for socket(), bind(), etc... */
#include <netdb.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>			/* Needed for exit() */
#include <stdlib.h>			/* Needed for atoi() */

#define  BUFFER 		2048	/* Allocate buffer for requests and responses */
#define  HEAD 			"HEAD / HTTP/1.0\r\nPragma: no-cache\r\nCache-Control: Max-age=0\r\n\r\n"
#define  DATE_LINE		"Date: "

static time_t mktime_from_rfc2616(const char *date) {
	struct tm timeinfo;
	time_t currtime = -1;
	char *monthinfo[12] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
	unsigned int i;

	/* Fill in values from RFC-compliant string. */
	timeinfo.tm_mday = atoi(date + 5);
	timeinfo.tm_year = atoi(date + 12) - 1900;
	timeinfo.tm_hour = atoi(date + 17);
	timeinfo.tm_min = atoi(date + 20);
	timeinfo.tm_sec = atoi(date + 23);

	/* Determine month. */
	timeinfo.tm_mon = -1;
	for (i = 0; i < (sizeof(monthinfo) / sizeof(monthinfo[0])); i++) {
		if (strncmp(date + 8, monthinfo[i], 3) == 0) {
			timeinfo.tm_mon = i;
			break;
		}
	}

	/* Create GMT value */
	currtime = mktime(&timeinfo);

	return(currtime);
}

static int set_clock(time_t timeval) {
	struct timeval tvinfo;
	int retval = -1;

	tvinfo.tv_sec = timeval;
	tvinfo.tv_usec = 0;

	retval = settimeofday(&tvinfo, NULL);
	printf ("Time set to: %s\n", ctime(&timeval));

	return(retval);
}

/* Calculate the difference between localtime and GMT/UT/UTC. */
static time_t get_gmtoffset(void) {
	static time_t retval = -1;
	time_t dummytime = 604800;
	struct tm *timeinfo = NULL;

	/* Return saved value. */
	if (retval != -1) {
		return(retval);
	}

	timeinfo = localtime(&dummytime);
	if (timeinfo == NULL) {
		return(-1);
	}
	if (timeinfo->tm_hour == 0) {
		timeinfo->tm_hour = 24;
	}
	retval = (timeinfo->tm_hour * 3600) + (timeinfo->tm_min * 60);

	timeinfo = gmtime(&dummytime);
	if (timeinfo == NULL) {
		retval = -1;
		return(-1);
	}
	if (timeinfo->tm_hour == 0) {
		timeinfo->tm_hour = 24;
	}
	retval -= (timeinfo->tm_hour * 3600) + (timeinfo->tm_min * 60);

	return(retval);
}

static time_t get_server_time(const char *host, unsigned int port) {
	unsigned int         server_s;		/* Server socket descriptor */
	struct sockaddr_in   server_addr;	/* Server Internet address */
	char                 out_buf[BUFFER];	/* Output buffer for HEAD request */
	char                 in_buf[BUFFER];	/* Input buffer for response */
	int 		     retcode;		/* Return code */

	char			remote_time[32];
	char			*pdate;
	struct			hostent	*hostinfo;
	time_t			remote_timeval = -1, retval = -1;

	hostinfo = gethostbyname(host);
	if (!hostinfo) {
		fprintf(stderr, "Host %s not found.\n", host);
		return(-1);
	}

	/* Create a socket */
	server_s = socket(AF_INET, SOCK_STREAM, 0);

	/* The web server socket's address information */
	server_addr.sin_family = AF_INET;
	server_addr.sin_port = htons(port);
	server_addr.sin_addr = *(struct in_addr *)*hostinfo -> h_addr_list;

	/* Do a connect (connect() blocks) */
	retcode = connect(server_s, (struct sockaddr *)&server_addr, sizeof(server_addr));
	if ( retcode < 0 ) {
		fprintf(stderr, "Connection to %s failed\n", host);
		return(-1);
	}

	/* Send the most simple HEAD request */
	strcpy(out_buf, HEAD);
	send(server_s, out_buf, strlen(out_buf), 0);

	/* Receive data from the web server */
	/* The return code from recv() is the number of bytes received */
	retcode = recv(server_s, in_buf, BUFFER, 0);
	if ( retcode <= 0 ) {
		fprintf(stderr, "No data received from %s\n", host);
		close(server_s);
		return(-1);
	}

	if ((pdate = (char *) strstr(in_buf, DATE_LINE)) != NULL) {
		strncpy(remote_time, pdate +  6, 29);

		remote_timeval = mktime_from_rfc2616(remote_time);

		retval = remote_timeval;
	} else {
		fprintf(stderr, "Date couldn't be retreived from %s\n", host);
	}

	close(server_s);

	return(retval);
}

static int compare_time(const void *a, const void *b) {
	const time_t *tmpptr = NULL;
	time_t  timea = 0, timeb = 0;

	tmpptr = a;
	timea = *tmpptr;
	tmpptr = b;
	timeb = *tmpptr;

	if (timea < timeb) {
		return(-1);
	}
	if (timea > timeb) {
		return(1);
	}

	return(0);
}

int main(int argc, char *argv[]) {
	time_t timevals[128] = {0}, timeval = -1;
	time_t mintime, maxtime = 0, avgtime = 0, meantime = 0;
	time_t stddevavgtime = 0;
	time_t tmptime = 0;
	time_t newtimeval = 0;
	time_t gmtoffset = 0;
	time_t start_time = 0, offset_time = 0;
	char *host = NULL, *portstr = NULL;
	int port = -1;
	int argind = 0, timeind = 0, meantimeind = 0, totaltimes = 0, totalstddevtimes = 0;
	int stddevtimes = 0;

	if ( argc <= 1 ) {
		printf ("Usage: htpdate <host>:[port] [<host>:[port] [...]]]\n");
		return(EXIT_FAILURE);
	}

	mintime = time(NULL) * 2;

	gmtoffset = get_gmtoffset();

	start_time = time(NULL);

	/* Process each argument as a host:port pair. */
	for (argind = 1; argind < argc; argind++) {
		host = argv[argind];

		portstr = strchr(host, ':');
		if (portstr != NULL) {
			/* Here, we modify argv.. good idea? */
			portstr[0] = '\0';

			portstr++;
			port = atoi(portstr);
			if (port <= 0) {
				fprintf(stderr, "Invalid port: %s.\n", portstr);
				continue;
			}
		} else {
			port = 80;
		}

		timeval = get_server_time(host, port);
		if (timeval < 0) {
			continue;
		}

		/* Remove the amount of time spent so far. */
		offset_time = time(NULL) - start_time;

		timevals[timeind++] = timeval - offset_time;
	}

	totaltimes = timeind;
	meantimeind = totaltimes / 2;

	if (totaltimes == 0) {
		fprintf(stderr, "No servers could be reached.\n");
		return(EXIT_FAILURE);
	}

	/* If there's only one time, don't bother with anything else. */
	if (totaltimes == 1) {
		newtimeval = timevals[0];
	} else {
		qsort(&timevals, totaltimes, sizeof(timevals[0]), compare_time);
		mintime = timevals[0];
		maxtime = timevals[totaltimes - 1];
		meantime = timevals[meantimeind];

		for (timeind = 0; timeind < totaltimes; timeind++) {
			tmptime = timevals[timeind];
			avgtime += tmptime - mintime; /* This trickery is to avoid overflowing the type. */
		}
		avgtime /= totaltimes;
		avgtime += mintime; /* Part of the same silly trickery. */

		/* If there are only 2 values, we can do no more. */
		if (totaltimes == 2) {
			newtimeval = avgtime;
		} else {
			printf("mintime = %li\n", mintime);
			printf("maxtime = %li\n", maxtime);
			printf("avgtime = %li\n", avgtime);
			printf("meantime = %li\n", meantime);

			stddevtimes = ((double) totaltimes) * 0.34; /* One standard deviation, +/- 34% about the mean. */

			for (timeind = (meantimeind - stddevtimes); timeind < (meantimeind + stddevtimes); timeind++) {
				tmptime = timevals[timeind];
				stddevavgtime += tmptime - mintime;
				totalstddevtimes++;
			}
			stddevavgtime /= totalstddevtimes;
			stddevavgtime += mintime;

			printf("stddevavgtime: %li (currtime = %li)\n", stddevavgtime, time(NULL));

			newtimeval = stddevavgtime;
		}
	}

	if (newtimeval <= 0) {
		fprintf(stderr, "Failed to calculate new time.\n");
		return(EXIT_FAILURE);
	}

	offset_time = time(NULL) - start_time;

	/* Add back the time this process took. */
	newtimeval += offset_time;

	set_clock(newtimeval + gmtoffset);

	return(EXIT_SUCCESS);
}
